/***************************************************************************
 *  Released into the public domain by Taj Morton 2007                     *
 *  tajmorton@gmail.com                                                    *
 *                                                                         *
 *  Do whatever you want to with this, but I'd appriciate it if you sent   *
 *  me the changes you make so I can merge them.                           *
 *                                                                         *
 *  Version: 1.0.1 2007-06-18 -- Add == and =(int) operators.              *
 *  Version 1.0:   2007-06-17                                              *
 ***************************************************************************/

#include <math.h>
#include <stdio.h>

#include <sstream> //for to_string

#include "decimal.h"

decimal::decimal( const int precision, const int value, RoundingMethod round )
{
	m_precision=precision;
	m_num=abs(value) * (long long int)pow(10,precision);
	m_round=round;
	
	if (value<0)
		m_is_positive=false;
	else
		m_is_positive=true;
}

decimal::decimal( const int precision, const double value, RoundingMethod round )
{
	m_round=round;
	setFromDouble(precision, value);
}

double decimal::to_double( ) const
{
	if (m_is_positive)
		return m_num/pow(10,m_precision);
	else {
		if (m_num!=0)
			return (m_num/pow(10,m_precision))*-1;
		else
			return 0.0;
	}
}

/**
 * The side with the greater precision is used for the
 * calculation. Then, the result is rounded to the left
 * hand side's precision setting.
 * @param rhs 
 * @return 
 */
decimal decimal::operator +( const decimal & rhs ) const
{
	decimal this_copy(*this);
	add(this_copy, rhs);
	return this_copy;
}

decimal decimal::operator +( const int & rhs ) const
{
	decimal this_copy(*this);
	add(this_copy, rhs);
	
	return this_copy;
}

decimal & decimal::operator ++( ) // this is the prefix operator
{
	add(*this, 1);
	return *this;
}

decimal decimal::operator ++( int ) // and this is the postfix operator
{
	decimal this_copy(*this);
	add(*this, 1);
	return this_copy;
}

decimal & decimal::operator +=( const decimal & rhs )
{
	add(*this, rhs);
	return *this;
}


decimal & decimal::operator +=( const int & rhs )
{
	add(*this, rhs);
	return *this;
}

decimal decimal::operator -( const decimal & rhs ) const
{
	decimal this_copy(*this);
	subtract(this_copy, rhs);
	
	return this_copy;
}

decimal decimal::operator -( const int & rhs ) const
{
	decimal this_copy(*this);
	subtract(this_copy, rhs);
	
	return this_copy;
}

decimal & decimal::operator -=( const decimal & rhs )
{
	subtract(*this, rhs);
	return *this;
}

decimal & decimal::operator -=( const int & rhs )
{
	subtract(*this, rhs);
	return *this;
}

decimal decimal::operator *( const decimal & rhs ) const
{
	decimal this_copy(*this);
	multiply(this_copy, rhs);
	
	return this_copy;
}

decimal decimal::operator *( const int & rhs ) const
{
	decimal this_copy(*this);
	multiply(this_copy, rhs);
	
	return this_copy;
}

decimal & decimal::operator *=( const decimal & rhs )
{
	multiply(*this, rhs);
	return *this;
}

decimal & decimal::operator *=( const int & rhs )
{
	multiply(*this, rhs);
	return *this;
}

decimal decimal::operator /( const decimal & rhs ) const
{
	decimal this_copy(*this);
	divide(this_copy, rhs);
	
	return this_copy;
}

decimal decimal::operator /( const int & rhs ) const
{
	decimal this_copy(*this);
	divide(this_copy, rhs);
	
	return this_copy;
}

decimal & decimal::operator /=( const decimal & rhs )
{
	divide(*this, rhs);
	return *this;
}

decimal & decimal::operator /=( const int & rhs )
{
	divide(*this, rhs);
	return *this;
}

/**
 * Changes precision of the value. If the current
 * precision is greater than the passed value, then
 * the value of the decimal is rounded according to the
 * round method currently set.
 * @param precision new precision
 */
void decimal::changePrecision( unsigned short int precision )
{
	if (precision>m_precision)
		m_num *= (long long int)pow(10,precision-m_precision);
	
	else if (precision<m_precision) {
		double newval = m_num * pow(10,precision-m_precision);
		
		double lastDigit = modf(newval, &newval);
		m_num=(long long int)newval;
		lastDigit *= 10;
		
		if (m_round==Normal) {
			if ((int)lastDigit > 4)
				m_num++;
		}
		else if (m_round==AllUp) {
			if ((int)lastDigit!=0) {
				m_num++;
			}
		}
	}
	
	m_precision=precision;
}

std::string decimal::to_string( ) const
{
	std::stringstream stream;
	stream << m_num;
	
	std::string str = stream.str();
	
	if ((str.length()!=m_precision) && ((m_num/(unsigned long long int)powl(10,m_precision))==0)) {
		// this handles the case where a decimal has no integer part (like 0.00003)
		// In this case, m_num is stored as 3, and is converted the same way
		// add the leading zeros:
		
		str.insert(0,std::string(m_precision, '0'));
	}
	
	if (str.length()-m_precision==0)
		str.insert(0,"0."); // we want a leading 0 if there is no integer part to this decimal
	else if (m_precision!=0)
		str.insert(str.length()-m_precision,".");
	
	if (!m_is_positive && (str.length()-m_precision!=0)) // don't add a "-" if we're a 0 number!
		str.insert(0,"-");
	
	return str;
}

std::ostream & operator <<( std::ostream & os, decimal d )
{
	os << d.to_string();
	return os;
}

decimal::operator std::string( ) const
{
	return to_string();
}

/**
 * Like the +, -, *, and / operators, these choose
 * the most accurate precision and compare based on that.
 * @param rhs 
 * @return 
 */
bool decimal::operator >( const decimal & rhs ) const
{
	if (m_is_positive && !rhs.m_is_positive)
		return true;
	else if (!m_is_positive && rhs.m_is_positive)
		return false;
	else {
		decimal this_copy(*this), rhs_copy(rhs);
		
		(m_precision>rhs.m_precision) ? rhs_copy.changePrecision(m_precision) : this_copy.changePrecision(rhs.m_precision);
		
		if (m_is_positive && rhs.m_is_positive)
			return this_copy.m_num>rhs_copy.m_num;
		else
			return this_copy.m_num<rhs_copy.m_num;
	}
}

bool decimal::operator >( const int & rhs ) const
{
	if (m_is_positive && rhs<0)
		return true;
	else if (!m_is_positive && rhs>0)
		return false;
	else {
		unsigned int rhs_scaled = abs(rhs)*(long long int)pow(10,m_precision);
	
		if (m_is_positive)
			return m_num>rhs_scaled;
		else
			return m_num<rhs_scaled;
	}
}

bool decimal::operator <( const decimal & rhs ) const
{
	if (m_is_positive && !rhs.m_is_positive)
		return false;
	else if (!m_is_positive && rhs.m_is_positive)
		return true;
	else {
		decimal this_copy(*this), rhs_copy(rhs);
		
		(m_precision>rhs.m_precision) ? rhs_copy.changePrecision(m_precision) : this_copy.changePrecision(rhs.m_precision);
		
		if (m_is_positive && rhs.m_is_positive)
			return this_copy.m_num<rhs_copy.m_num;
		else
			return this_copy.m_num>rhs_copy.m_num;
	}
}

bool decimal::operator <( const int & rhs ) const
{
	if (m_is_positive && rhs<0)
		return false;
	else if (!m_is_positive && rhs>0)
		return true;
	else {
		unsigned int rhs_scaled = abs(rhs)*(long long int)pow(10,m_precision);
	
		if (m_is_positive)
			return m_num<rhs_scaled;
		else
			return m_num>rhs_scaled;
	}
}

void decimal::setFromDouble( const int precision, const double value )
{
	double fractpart, intpart;
	m_precision=precision;
	
	fractpart = modf(fabs(value), &intpart);
	m_num = (long long int)intpart * (long long int)pow(10,precision);
	
	// move decimal point over so now fractpart's `precision` digits
	// are on the int side
	fractpart *= pow(10,precision);
	
	// combine fracpart into (already shifted) m_intpart
	double roundCheck = modf(fractpart, &fractpart);
	m_num+=(long long int)fractpart;
	
	double nextDigit;
	roundCheck = modf(roundCheck*10, &nextDigit);
	
	if (m_round==Normal) {
		if ((int)nextDigit > 4)
			m_num++;
	}
	else if (m_round==AllUp) {
		if ((int)nextDigit!=0) {
			m_num++;
		}
	}
	
	if (value<0)
		m_is_positive=false;
	else
		m_is_positive=true;
}

decimal & decimal::operator =( const double & rhs )
{
	setFromDouble(m_precision,rhs);
	return *this;
}

void decimal::add( decimal & dec, const decimal & value )
{
	int old_dec_precision=dec.m_precision;
	decimal value_copy(value);
	
	(dec.m_precision>value_copy.m_precision) ? value_copy.changePrecision(dec.m_precision) : dec.changePrecision(value_copy.m_precision);
	
	if (dec.m_is_positive==value_copy.m_is_positive)
		dec.m_num+=value_copy.m_num;
	else {
		if (dec.m_num > value_copy.m_num)
			dec.m_num-=value_copy.m_num;
		else if (dec.m_num < value_copy.m_num) {
			dec.m_num=value_copy.m_num-dec.m_num;
			dec.m_is_positive=false;
		}
		else {
			dec.m_num=0;
		}
			
	}
	
	dec.changePrecision(old_dec_precision);
}

void decimal::add( decimal & dec, const int & value )
{
	unsigned long long int rhs_scaled=abs(value * (long long int)pow(10,dec.m_precision));
	
	if ((value>0 && dec.m_is_positive) || (value<0 && !dec.m_is_positive)) {
		dec.m_num += rhs_scaled;
	}
	else {
		// negative int, special case so we don't overflow (since m_num is unsigned)
		if (dec.m_num > rhs_scaled)
			dec.m_num-=rhs_scaled;
		else if (dec.m_num < rhs_scaled) {
			dec.m_num=rhs_scaled-dec.m_num;
			dec.m_is_positive=false;
		}
		else
			dec.m_num=0;
	}
}

void decimal::subtract( decimal & dec, const decimal & value )
{
	int old_dec_precision=dec.m_precision;
	decimal value_copy(value);
	
	(dec.m_precision>value_copy.m_precision) ? value_copy.changePrecision(dec.m_precision) : dec.changePrecision(value_copy.m_precision);
	
	if (dec.m_is_positive != value_copy.m_is_positive) {
		dec.m_num+=value_copy.m_num;
		
		if (value_copy.m_is_positive)
			dec.m_is_positive=false;
	}
	else if (!dec.m_is_positive && !value_copy.m_is_positive) {
		if (dec.m_num > value_copy.m_num) {
			dec.m_num-=value_copy.m_num;
			dec.m_is_positive=false;
		}
		else if (dec.m_num < value_copy.m_num) {
			dec.m_num=value_copy.m_num-dec.m_num;
			dec.m_is_positive=true;
		}
		else {
			dec.m_num=0;
		}
	}
	else { // both are positive
		if (dec.m_num > value_copy.m_num) {
			dec.m_num-=value_copy.m_num;
		}
		else if (dec.m_num < value_copy.m_num) {
			dec.m_num=value_copy.m_num-dec.m_num;
			dec.m_is_positive=false;
		}
		else {
			dec.m_num=0;
		}
	}
	
	dec.changePrecision(old_dec_precision);
}

void decimal::subtract( decimal & dec, const int & value )
{
	unsigned long long int value_scaled = abs(value * (long long int)pow(10,dec.m_precision));
	
	if ((dec.m_is_positive && value<0) || (!dec.m_is_positive && value>0)) {
		dec.m_num+=value_scaled;
		
		if (value>0)
			dec.m_is_positive=false;
	}
	else if (!dec.m_is_positive && value<0) {
		if (dec.m_num > value_scaled) {
			dec.m_num-=value_scaled;
			dec.m_is_positive=false;
		}
		else if (dec.m_num < value_scaled) {
			dec.m_num=value_scaled-dec.m_num;
			dec.m_is_positive=true;
		}
		else {
			dec.m_num=0;
		}
	}
	else { // both are positive
		if (dec.m_num > value_scaled) {
			dec.m_num-=value_scaled;
		}
		else if (dec.m_num < value_scaled) {
			dec.m_num=value_scaled-dec.m_num;
			dec.m_is_positive=false;
		}
		else {
			dec.m_num=0;
		}
	}
}

void decimal::multiply( decimal & dec, const decimal & value )
{
	dec.m_num*=value.m_num;
	
	// place the decimal point in the new position
	int old_precision=dec.m_precision;
	dec.m_precision=value.m_precision+dec.m_precision;
	
	// and round to the old position
	dec.changePrecision(old_precision);
	
	if (dec.m_is_positive == value.m_is_positive)
		dec.m_is_positive=true;
	else
		dec.m_is_positive=false;
}

void decimal::multiply( decimal & dec, const int & value )
{
	dec.m_num *= abs(value);
	
	if ((dec.m_is_positive && value>0) || (!dec.m_is_positive && value<0))
		dec.m_is_positive=true;
	else
		dec.m_is_positive=false;
}

void decimal::divide( decimal & dec, const decimal & value )
{
	decimal value_copy(value);
	int old_precision=dec.m_precision;
	
	(dec.m_precision>value.m_precision) ? value_copy.changePrecision(dec.m_precision) : dec.changePrecision(value.m_precision);
	
	double div_result = (double)dec.m_num / (double)value_copy.m_num;
	
	double intpart, fractpart;
	fractpart = modf(div_result, &intpart);
	
	intpart *= pow(10,dec.m_precision); //make room for the fractpart
	fractpart *= pow(10,dec.m_precision); //move as much as we want over to the int side of the decimal point
	
	// combine fracpart into (already shifted) m_intpart
	dec.m_num = (long long int)intpart + (long long int)fractpart;
	
	double roundCheck = modf(fractpart, &fractpart);
	
	double nextDigit;
	roundCheck = modf(roundCheck*10, &nextDigit);
	
	if (dec.m_round==Normal) {
		if ((int)nextDigit > 4)
			dec.m_num++;
	}
	else if (dec.m_round==AllUp) {
		if ((int)nextDigit!=0) {
			dec.m_num++;
		}
	}
	
	if (dec.m_is_positive == value_copy.m_is_positive)
		dec.m_is_positive=true;
	else
		dec.m_is_positive=false;
	
	dec.changePrecision(old_precision);
}

void decimal::divide( decimal & dec, const int & value )
{
	dec.m_num /= abs(value);
	
	if ((dec.m_is_positive && value>0) || (!dec.m_is_positive && value<0))
		dec.m_is_positive=true;
	else
		dec.m_is_positive=false;
}

bool decimal::operator ==( const decimal & rhs ) const
{
	if (m_is_positive != rhs.m_is_positive)
		return false;
	
	if (m_precision != rhs.m_precision) {
		decimal this_copy(*this), rhs_copy(rhs);
		
		(this_copy.m_precision>rhs_copy.m_precision) ? rhs_copy.changePrecision(this_copy.m_precision) : this_copy.changePrecision(rhs_copy.m_precision);
		
		return this_copy.m_num==rhs_copy.m_num;
	}
	else
		return m_num==rhs.m_num;
}

bool decimal::operator ==( const int & rhs ) const
{
	if (m_is_positive && rhs<0)
		return false;
	else if (!m_is_positive && rhs>0)
		return false;
	else {
		unsigned long long int rhs_scaled = abs(rhs * (long long int)pow(10,m_precision));
		
		return m_num==rhs_scaled;
	}
}

void decimal::operator =( const int & rhs )
{
	m_num=abs(rhs * (long long int)pow(10,m_precision));
	if (rhs<0)
		m_is_positive=false;
	else
		m_is_positive=true;
}


